import { AudioPlayerStatus, createAudioPlayer, createAudioResource, entersState, getVoiceConnection, joinVoiceChannel, StreamType, VoiceConnectionStatus, } from "@discordjs/voice";
import { StageChannel } from "discord.js";
import { getAllAudioBase64 } from "google-tts-api";
import ytdl from "ytdl-core";
import { sLogger } from "../logger";
import { shorten } from "../utils/string";
export var VoiceValidateResult;
(function (VoiceValidateResult) {
    VoiceValidateResult[VoiceValidateResult["OK"] = 0] = "OK";
    VoiceValidateResult[VoiceValidateResult["NO_CHANNEL"] = 1] = "NO_CHANNEL";
    VoiceValidateResult[VoiceValidateResult["STAGE_CHANNEL"] = 2] = "STAGE_CHANNEL";
})(VoiceValidateResult || (VoiceValidateResult = {}));
export class VoiceControl {
    constructor(msg, handleDisconnect) {
        this.speakQueue = [];
        this.handleDisconnect = handleDisconnect;
        const validateResult = VoiceControl.validateUser(msg);
        if (validateResult)
            throw validateResult;
        this.player = createAudioPlayer();
        const voiceChannel = msg.member.voice.channel;
        this.guildID = voiceChannel.guild.id;
        this.channelName = voiceChannel.name;
        this.voiceChannelID = voiceChannel.id;
        const connection = joinVoiceChannel({
            channelId: voiceChannel.id,
            guildId: voiceChannel.guild.id,
            adapterCreator: voiceChannel.guild
                .voiceAdapterCreator,
            selfMute: false,
        });
        connection.subscribe(this.player);
        // * https://discordjs.guide/voice/voice-connections.html#handling-disconnects
        connection.on(VoiceConnectionStatus.Disconnected, async (oldState, newState) => {
            try {
                await Promise.race([
                    entersState(connection, VoiceConnectionStatus.Signalling, 5000),
                    entersState(connection, VoiceConnectionStatus.Connecting, 5000),
                ]);
                // Seems to be reconnecting to a new channel - ignore disconnect
            }
            catch (error) {
                // Seems to be a real disconnect which SHOULDN'T be recovered from
                this.destruct();
            }
        });
    }
    isSameChannel(channel) {
        return channel?.id == this.voiceChannelID;
    }
    static validateUser(msg) {
        const channel = msg.member?.voice.channel;
        // * This Bot and every others cannot enter *undefined* channel
        if (!channel)
            return VoiceValidateResult.NO_CHANNEL;
        // * This Bot doesn't support Stage Channel
        if (channel instanceof StageChannel)
            return VoiceValidateResult.STAGE_CHANNEL;
        return VoiceValidateResult.OK;
    }
    destruct() {
        const connection = getVoiceConnection(this.guildID);
        connection?.destroy();
        this.handleDisconnect();
    }
    async waitTillReady() {
        const connection = getVoiceConnection(this.guildID);
        if (!connection)
            return false;
        try {
            await entersState(connection, VoiceConnectionStatus.Ready, 5000);
            sLogger.log(`Successfully joined ${this.channelName} as SOD Mode`, "SUCCESS");
            return true;
        }
        catch (err) {
            sLogger.log(`Error joining ${this.channelName} : ${err}`, "ERROR");
            return false;
        }
    }
    async speak(content) {
        sLogger.log(`[TTS] Started Speaking ${shorten(content, 20)} to ${this.channelName}`);
        // TODO Optimize by getAllAudioUrl then get each Audio async
        const results = await getAllAudioBase64(content, {
            lang: "th",
            slow: false,
            host: "https://translate.google.com",
        });
        this.speakQueue = results.map((res) => createAudioResource("data:audio/ogg;base64," + res.base64, {
            inputType: StreamType.OggOpus,
        }));
        this.player.play(this.speakQueue[0]);
        this.speakQueue.shift();
        this.player.on(AudioPlayerStatus.Idle, (() => {
            if (this.speakQueue.length) {
                this.player.play(this.speakQueue.shift());
            }
        }).bind(this));
    }
    async playSong(url) {
        // https://stackoverflow.com/questions/63199238/discord-js-ytdl-error-input-stream-status-code-416
        const musicStream = ytdl(url, {
            quality: "highestaudio",
            highWaterMark: 1 << 25,
        });
        const musicRc = createAudioResource(musicStream);
        this.player.play(musicRc);
    }
}
